#include"POBox.h"
#include"POMessage.h"
#include"PostOffice.h"
#include<iostream>
#include<unistd.h>
#include<stdarg.h>
#include<Tracer.h>
#include<POGroupMsg.h>
#include<POTextMsg.h>
#include<POErrorMsg.h>
#include<qapplication.h>
using namespace std;


//=============================================================================
// Function Name: POBox
///     Constructor
/// \param name - the name of the inbox
/// \param po - a pointer to the post office
//=============================================================================
POBox::POBox( const string& name, PostOffice* po )
:QObject(), m_name(name), m_po(po), m_queuestate(true)
{
}

//=============================================================================
// Function Name: ~POBox
///     Delete all unread messages
//=============================================================================
POBox::~POBox()
{
    list<const POMessage*>::iterator itr = m_inbox.begin();
    while( itr != m_inbox.end() )
    {
        delete *itr;
        itr++;
    }
}

//=============================================================================
// Function Name: popMessage
///     Returns the oldest message on the stack, you are responsible for 
///     deleting it.
/// \return a pointer
//=============================================================================
const POMessage* POBox::popMessage()
{
    const POMessage* back = m_inbox.back();
    m_inbox.pop_back();
    return back;
}

//=============================================================================
// Function Name: HasMessages
///     Returns true if the mail box contains messages
/// \param check_now - true if you want Pony Express to check sockets before
///                    before checking messages.  This will happen automatically
///                    next time qApp->processEvents() is called, but for the 
///                    lowest latency we can call it now
/// \return true if there are messages, otherwise false
//=============================================================================
bool POBox::hasMessages( bool check_now )const
{
    // if check, it will update the socket for the latest messages
    if( check_now ) qApp->processEvents();
    return m_inbox.size() > 0;
}

//=============================================================================
// Function Name: SendMessage
///     Sets the message from field to POBox::GetName()
/// \param msg - the message to send
/// \param adr - the address to send it to
/// \param now - wether or not to write it out the socket now, or wait until
///              the next cycle
/// \return true - if the message could be sent, otherwise false
//=============================================================================
bool POBox::sendMessage( POMessage* msg, const POAddress& adr, bool now )
{
/*    Tracer::debug( "POBox<%s>::sendMessage( %s, %s.%s )", 
                    m_name.c_str(),
                    msg->messageType().c_str(),
                    adr.getPO().c_str(),
                    adr.getBox().c_str() );
*/
    msg->setDestination( adr );
    msg->setSource( POAddress( m_po->getName(), m_name ) );
    emit messageSent( msg );
    bool result = m_po->sendMessage( msg );
    
    // send the message out the socket now
    if( now ) 
        qApp->processEvents();
    return result;
}


//=============================================================================
// Function Name: sendBlockingMessage
///     Sends a message and blocks until a message is sent in response or 
///     the message times out.  On a timeout or other error a 
///     POErrorMsg* exception is thrown.
/// \param msg - the message to send
/// \param adr - the address to send the message to
/// \param timeout - the max time to wait for a response, default is 2 seconds
/// \return the message returned 
//=============================================================================
const POMessage* POBox::sendBlockingMessage( POMessage* msg, 
                                            const POAddress& adr,
                                            const DTime& timeout )
{
    int tag = DTime::current().usec();
    msg->setDestination( adr );
    msg->setSource( POAddress( m_po->getName(), m_name ) );
    msg->setTag( tag );
    emit messageSent( msg );
    m_po->sendMessage( msg );
    DTime timer;
    timer.start();
    while( timer.elapsed() < timeout )
    {
        qApp->processEvents();
        list<const POMessage*>::iterator itr = m_inbox.begin();
        while( itr != m_inbox.end() )
        {
            if( (*itr)->getTag() == tag )
            {
                const POMessage* msg = *itr;
                m_inbox.erase(itr);
                return msg;
            }
            itr++;
        }
        usleep( 1 );
    }
    throw new POErrorMsg( "sendBlockingMessage( msg, %s.%s, ... )\n"
                          "\tTimeout", adr.getPO().c_str(), adr.getBox().c_str() );
    return NULL;
}


/**
    Sends a text message to the specified POBox name on the local POPostOffice
    @param boxname - the name of the pobox to send the message to
    @param format - the text to send
    @return true if adr is a valid address
*/
bool POBox::sendTextMsg( const string& boxname, const char* format, ... )
{
    char* temp = new char[5000];
    va_list ap;
    va_start( ap, format );
    vsnprintf( temp, 4999, format, ap );
    va_end(ap);
    temp[4999] = '\0';
    POTextMsg msg( temp );
    delete[] temp;
    return sendMessage( &msg, POAddress( m_po->getName(), boxname ) );
}

//=============================================================================
// Function Name: sendTextMsg
///     Provided for convience, sends a POTextMsg to the specified address
/// \param adr - the address to send the message to
/// \param format - the text to send
/// \return true if adr is a valid address
//=============================================================================
bool POBox::sendTextMsg( const POAddress& adr, const char* format, ... )
{
    char* temp = new char[5000];
    va_list ap;
    va_start( ap, format );
    vsnprintf( temp, 4999, format, ap );
    va_end(ap);
    temp[4999] = '\0';
    POTextMsg msg( temp );
    delete[] temp;
    return sendMessage( &msg, adr );
}


//=============================================================================
// Function Name: AddMessage
///     Used by the PostOffice to add incomming messages to the box and then
///     emits MessageArrived()
/// \param msg - the incomming message
//=============================================================================
void POBox::addMessage( const POMessage* msg )
{
    Tracer::debug( 7, "POBox<%s>::addMessage( %s, %s.%s )", 
                    getName().c_str(),
                    msg->messageType().c_str(),
                    msg->getSource().getPO().c_str(),
                    msg->getSource().getBox().c_str() );
    if( m_queuestate )
    {
        m_inbox.push_back( msg );
        emit messageArrived();
    }
    emit messageArrived(msg);

    if( !m_queuestate ) delete msg;
}


//=============================================================================
// Function Name: joinGroup
///     Sends a message to the group that will add this POBox to the list of
///     group members.  This box will start receiving all messages sent to 
///     the group
/// \param groupadr - the address of the group to join
/// \param create   - ignored for the time being
/// \throws POErrorMsg* if an error occured
//=============================================================================
void POBox::joinGroup( const POAddress& groupadr, bool create )
{
    POGroupMsg joinmsg(POGroupMsg::JOIN);
    sendMessage( &joinmsg, groupadr );
}

//=============================================================================
// Function Name: joinGroup
///     Sends a message to the group that will add this POBox to the list of
///     group members.  This box will start receiving all messages sent to 
///     the group
/// \param groupname - the POBox of the group, assumes local Post Office 
/// \param create   - ignored for the time being
/// \throws POErrorMsg* if an error occured
//=============================================================================
void POBox::joinGroup( const string& groupname, bool create )
{
    joinGroup( POAddress( m_po->getName(), groupname ), create );
}

//=============================================================================
// Function Name: leaveGroup
///     Sends a message to the group that will remove this POBox from the list 
///     of group members.  This box will stop receiving all messages sent to 
///     the group
/// \param groupadr - the address of the group to join
/// \throws POErrorMsg* if an error occured
//=============================================================================
void POBox::leaveGroup( const POAddress& groupadr )
{
    POGroupMsg leavemsg(POGroupMsg::LEAVE);
    sendMessage( &leavemsg, groupadr );
}

//=============================================================================
// Function Name: leaveGroup
///     Sends a message to the group that will remove this POBox from the list 
///     of group members.  This box will stop receiving all messages sent to 
///     the group
/// \param groupname - the name of the group, assumes the local Post Office
/// \throws POErrorMsg* if an error occured
//=============================================================================
void POBox::leaveGroup( const string& groupname )
{
    POGroupMsg leavemsg(POGroupMsg::LEAVE);
    sendMessage( &leavemsg, POAddress( m_po->getName(), groupname ) );
}
