#ifndef _POINTERREF_H_
#define _POINTERREF_H_

namespace Diverse
{
/** @class PointerRef
    @brief Keeps track of references to a pointer
    PointerRef keeps track of the number of times a given pointer is
    copied.  When no one has a handle to the pointer it is automatically deleted.
*/
template<class type>
class PointerRef 
{
    public:
        typedef type* typeptr;
        PointerRef() {m_refs = NULL; m_data = new type(); inc(); }
        /// @param the pointer to reference, 
        ///     make sure it is dynamically alocated
        PointerRef( type* data );
        PointerRef( const PointerRef<type>& cpy );
        ~PointerRef();

        PointerRef<type>& operator = ( const PointerRef<type>& cyp );
        PointerRef<type>& operator = ( type* cpy ) { set(cpy); return *this; }
//        operator type()    { return *m_data; }
        operator typeptr() { return m_data; }

        
        type&             operator*()       {  return *m_data;    }
        type*             operator->()const {  return  m_data;    }
        
        /// Be very careful not to store this pointer as the object
        /// will be automatically deleted at some point
        /// @return the pointer
        type* ptr()const        { return m_data; }

        /// Used to set the data pointer, decriments the reference
        /// count on the old data before setting it for the new data.
        void  set( type* data ); 

        void deref();
        void inc();

        int numrefs()const { return m_refs?(*m_refs):0; }

    private:
        int*  m_refs;
        type* m_data;
};

template<class type>
PointerRef<type>::PointerRef( type* data )
{
    m_refs = NULL;
    m_data = NULL;
    set( data );
}


template<class type>
PointerRef<type>::PointerRef( const PointerRef<type>& cpy )
{
    m_refs = cpy.m_refs;
    m_data = cpy.m_data;
    inc();
}
template<class type>
PointerRef<type>::~PointerRef()
{
    deref();
}

template<class type>
void PointerRef<type>::set( type* data )
{
    deref();
    inc();
    m_data = data;
}

template<class type>
void PointerRef<type>::deref()
{
    if( m_refs == NULL ) return;
    (*m_refs) -= 1;
    if( (*m_refs) == 0 )
    {
        delete m_refs;
        if( m_data )
            delete m_data;
        m_refs = NULL;
        m_data = NULL;
    }
}
template<class type>
void PointerRef<type>::inc()
{
    if( m_refs == NULL ) 
        m_refs = new int(0);
    (*m_refs) += 1;
}
template<class type>
PointerRef<type>& PointerRef<type>::operator=( const PointerRef<type>& cpy )
{
    // Don't do anything if it is the same object
    if( this == &cpy ) return *this;
 
    // Decrement Ref Count on the current data and delete if necessary
    deref();
    
    // Assign the data
    m_refs = cpy.m_refs;
    m_data = cpy.m_data;

    // Inc ref count
    inc();
    return *this;
}

}
#endif
